"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getInstructionFunctionHighLevelFragment = void 0;
const shared_1 = require("../../../shared");
const asyncHelpers_1 = require("../asyncHelpers");
const common_1 = require("./common");
const instructionByteDelta_1 = require("./instructionByteDelta");
const instructionInputResolved_1 = require("./instructionInputResolved");
const instructionInputType_1 = require("./instructionInputType");
const instructionRemainingAccounts_1 = require("./instructionRemainingAccounts");
function getInstructionFunctionHighLevelFragment(scope) {
    var _a;
    const { useAsync, instructionNode, programNode, resolvedInputs, renamedArgs, dataArgsManifest, asyncResolvers, nameApi, customInstructionData, } = scope;
    if (useAsync &&
        !(0, asyncHelpers_1.hasAsyncFunction)(instructionNode, resolvedInputs, asyncResolvers)) {
        return (0, common_1.fragment)('');
    }
    const customData = customInstructionData.get(instructionNode.name);
    const hasAccounts = instructionNode.accounts.length > 0;
    const hasDataArgs = !!customData ||
        instructionNode.arguments.filter((field) => !field.defaultValue || field.defaultValueStrategy !== 'omitted').length > 0;
    const hasExtraArgs = ((_a = instructionNode.extraArguments) !== null && _a !== void 0 ? _a : []).filter((field) => !field.defaultValue || field.defaultValueStrategy !== 'omitted').length > 0;
    const hasAnyArgs = hasDataArgs || hasExtraArgs;
    const instructionDataName = nameApi.instructionDataType(instructionNode.name);
    const argsTypeFragment = (0, common_1.fragment)(customData
        ? dataArgsManifest.looseType.render
        : nameApi.dataArgsType(instructionDataName));
    if (customData) {
        argsTypeFragment.mergeImportsWith(dataArgsManifest.looseType);
    }
    const functionName = useAsync
        ? nameApi.instructionAsyncFunction(instructionNode.name)
        : nameApi.instructionSyncFunction(instructionNode.name);
    const lowLevelFunctionName = nameApi.instructionRawFunction(instructionNode.name);
    const typeParamsFragment = getTypeParams(instructionNode, programNode);
    const instructionTypeFragment = getInstructionType({
        ...scope,
        withSigners: false,
    });
    const instructionTypeWithSignersFragment = getInstructionType({
        ...scope,
        withSigners: true,
    });
    // Input.
    const inputTypeFragment = (0, instructionInputType_1.getInstructionInputTypeFragment)({
        ...scope,
        withSigners: false,
    });
    const inputTypeWithSignersFragment = (0, instructionInputType_1.getInstructionInputTypeFragment)({
        ...scope,
        withSigners: true,
    });
    const inputTypeCallFragment = getInputTypeCall({
        ...scope,
        withSigners: false,
    });
    const inputTypeCallWithSignersFragment = getInputTypeCall({
        ...scope,
        withSigners: true,
    });
    const renamedArgsText = [...renamedArgs.entries()]
        .map(([k, v]) => `${k}: input.${v}`)
        .join(', ');
    const resolvedInputsFragment = (0, instructionInputResolved_1.getInstructionInputResolvedFragment)(scope);
    const remainingAccountsFragment = (0, instructionRemainingAccounts_1.getInstructionRemainingAccountsFragment)(scope);
    const byteDeltaFragment = (0, instructionByteDelta_1.getInstructionByteDeltaFragment)(scope);
    const resolvedFragment = (0, common_1.mergeFragments)([resolvedInputsFragment, remainingAccountsFragment, byteDeltaFragment], (renders) => renders.join('\n\n'));
    const hasRemainingAccounts = remainingAccountsFragment.render !== '';
    const hasByteDeltas = byteDeltaFragment.render !== '';
    const hasResolver = resolvedFragment.hasFeatures('instruction:resolverScopeVariable');
    const getReturnType = (instructionType) => {
        let returnType = instructionType;
        if (hasByteDeltas) {
            returnType = `${returnType} & IInstructionWithByteDelta`;
        }
        return useAsync ? `Promise<${returnType}>` : returnType;
    };
    const functionFragment = (0, common_1.fragmentFromTemplate)('instructionFunctionHighLevel.njk', {
        instruction: instructionNode,
        program: programNode,
        hasAccounts,
        hasDataArgs,
        hasExtraArgs,
        hasAnyArgs,
        argsTypeFragment,
        functionName,
        lowLevelFunctionName,
        typeParamsFragment,
        instructionTypeFragment,
        instructionTypeWithSignersFragment,
        inputTypeFragment,
        inputTypeWithSignersFragment,
        inputTypeCallFragment,
        inputTypeCallWithSignersFragment,
        renamedArgs: renamedArgsText,
        resolvedFragment,
        hasRemainingAccounts,
        hasByteDeltas,
        hasResolver,
        useAsync,
        getReturnType,
    })
        .mergeImportsWith(typeParamsFragment, instructionTypeFragment, instructionTypeWithSignersFragment, inputTypeFragment, inputTypeWithSignersFragment, inputTypeCallFragment, inputTypeCallWithSignersFragment, resolvedFragment, argsTypeFragment)
        .addImports('solanaAddresses', ['Address']);
    if (hasAccounts) {
        functionFragment
            .addImports('solanaInstructions', ['IAccountMeta'])
            .addImports('shared', ['getAccountMetasWithSigners', 'ResolvedAccount']);
    }
    if (hasByteDeltas) {
        functionFragment.addImports('shared', ['IInstructionWithByteDelta']);
    }
    return functionFragment;
}
exports.getInstructionFunctionHighLevelFragment = getInstructionFunctionHighLevelFragment;
function getTypeParams(instructionNode, programNode) {
    const typeParams = [
        ...instructionNode.accounts.map((account) => `TAccount${(0, shared_1.pascalCase)(account.name)} extends string`),
        `TProgram extends string = "${programNode.publicKey}"`,
    ];
    return (0, common_1.fragment)(typeParams.filter((x) => !!x).join(', '));
}
function getInstructionType(scope) {
    const { instructionNode, withSigners, nameApi } = scope;
    const instructionTypeName = withSigners
        ? nameApi.instructionWithSignersType(instructionNode.name)
        : nameApi.instructionType(instructionNode.name);
    const accountTypeParamsFragments = instructionNode.accounts.map((account) => {
        const typeParam = `TAccount${(0, shared_1.pascalCase)(account.name)}`;
        const camelName = (0, shared_1.camelCase)(account.name);
        if (account.isSigner === 'either' && withSigners) {
            const signerRole = account.isWritable
                ? 'WritableSignerAccount'
                : 'ReadonlySignerAccount';
            return (0, common_1.fragment)(`typeof input["${camelName}"] extends TransactionSigner<${typeParam}> ` +
                `? ${signerRole}<${typeParam}> & IAccountSignerMeta<${typeParam}> ` +
                `: ${typeParam}`)
                .addImports('solanaInstructions', [signerRole])
                .addImports('solanaSigners', ['IAccountSignerMeta']);
        }
        return (0, common_1.fragment)(typeParam);
    });
    return (0, common_1.mergeFragments)([(0, common_1.fragment)('TProgram'), ...accountTypeParamsFragments], (renders) => renders.join(', ')).mapRender((r) => `${instructionTypeName}<${r}>`);
}
function getInputTypeCall(scope) {
    const { instructionNode, withSigners, useAsync, nameApi } = scope;
    const syncInputTypeName = withSigners
        ? nameApi.instructionSyncInputWithSignersType(instructionNode.name)
        : nameApi.instructionSyncInputType(instructionNode.name);
    const asyncInputTypeName = withSigners
        ? nameApi.instructionAsyncInputWithSignersType(instructionNode.name)
        : nameApi.instructionAsyncInputType(instructionNode.name);
    const inputTypeName = useAsync ? asyncInputTypeName : syncInputTypeName;
    if (instructionNode.accounts.length === 0)
        return (0, common_1.fragment)(inputTypeName);
    const accountTypeParams = instructionNode.accounts
        .map((account) => `TAccount${(0, shared_1.pascalCase)(account.name)}`)
        .join(', ');
    return (0, common_1.fragment)(`${inputTypeName}<${accountTypeParams}>`);
}
//# sourceMappingURL=instructionFunctionHighLevel.js.map