"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getInstructionInputTypeFragment = void 0;
const nodes_1 = require("../../../nodes");
const shared_1 = require("../../../shared");
const ImportMap_1 = require("../ImportMap");
const asyncHelpers_1 = require("../asyncHelpers");
const common_1 = require("./common");
function getInstructionInputTypeFragment(scope) {
    var _a;
    const { instructionNode, resolvedInputs, renamedArgs, dataArgsManifest, programNode, withSigners, asyncResolvers, useAsync, nameApi, customInstructionData, } = scope;
    // Accounts.
    const accountImports = new ImportMap_1.ImportMap();
    const accounts = instructionNode.accounts.map((account) => {
        const typeParam = `TAccount${(0, shared_1.pascalCase)(account.name)}`;
        const resolvedAccount = resolvedInputs.find((input) => input.kind === 'instructionAccountNode' && input.name === account.name);
        const hasDefaultValue = !!resolvedAccount.defaultValue &&
            !(0, nodes_1.isNode)(resolvedAccount.defaultValue, [
                'identityValueNode',
                'payerValueNode',
            ]) &&
            (useAsync ||
                !(0, asyncHelpers_1.isAsyncDefaultValue)(resolvedAccount.defaultValue, asyncResolvers));
        const type = getAccountType(resolvedAccount, withSigners);
        accountImports.mergeWith(type);
        return {
            ...resolvedAccount,
            typeParam,
            optionalSign: hasDefaultValue || resolvedAccount.isOptional ? '?' : '',
            type: type.render,
        };
    });
    // Arg link imports.
    const customData = customInstructionData.get(instructionNode.name);
    const argLinkImports = new ImportMap_1.ImportMap();
    if (customData) {
        argLinkImports.mergeWith(dataArgsManifest.looseType);
    }
    // Arguments.
    const resolveArg = (arg) => {
        var _a;
        const resolvedArg = resolvedInputs.find((input) => (0, nodes_1.isNode)(input, 'instructionArgumentNode') && input.name === arg.name);
        if (arg.defaultValue && arg.defaultValueStrategy === 'omitted')
            return [];
        const renamedName = (_a = renamedArgs.get(arg.name)) !== null && _a !== void 0 ? _a : arg.name;
        const optionalSign = arg.defaultValue || (resolvedArg === null || resolvedArg === void 0 ? void 0 : resolvedArg.defaultValue) ? '?' : '';
        return [
            {
                ...arg,
                ...resolvedArg,
                renamedName,
                optionalSign,
            },
        ];
    };
    const instructionDataName = nameApi.instructionDataType(instructionNode.name);
    const instructionExtraName = nameApi.instructionExtraType(instructionNode.name);
    const dataArgsType = customData
        ? nameApi.dataArgsType(customData.importAs)
        : nameApi.dataArgsType(instructionDataName);
    const dataArgs = customData
        ? []
        : instructionNode.arguments.flatMap(resolveArg);
    const extraArgsType = nameApi.dataArgsType(instructionExtraName);
    const extraArgs = ((_a = instructionNode.extraArguments) !== null && _a !== void 0 ? _a : []).flatMap(resolveArg);
    const syncInputType = withSigners
        ? nameApi.instructionSyncInputWithSignersType(instructionNode.name)
        : nameApi.instructionSyncInputType(instructionNode.name);
    const asyncInputType = withSigners
        ? nameApi.instructionAsyncInputWithSignersType(instructionNode.name)
        : nameApi.instructionAsyncInputType(instructionNode.name);
    const instructionInputType = useAsync ? asyncInputType : syncInputType;
    return (0, common_1.fragmentFromTemplate)('instructionInputType.njk', {
        instruction: instructionNode,
        program: programNode,
        instructionInputType,
        accounts,
        dataArgs,
        dataArgsType,
        extraArgs,
        extraArgsType,
    })
        .mergeImportsWith(accountImports, argLinkImports)
        .addImports('solanaAddresses', ['Address']);
}
exports.getInstructionInputTypeFragment = getInstructionInputTypeFragment;
function getAccountType(account, withSigners) {
    const typeParam = `TAccount${(0, shared_1.pascalCase)(account.name)}`;
    if (withSigners) {
        if (account.isPda && account.isSigner === false) {
            return (0, common_1.fragment)(`ProgramDerivedAddress<${typeParam}>`).addImports('solanaAddresses', ['ProgramDerivedAddress']);
        }
        if (account.isPda && account.isSigner === 'either') {
            return (0, common_1.fragment)(`ProgramDerivedAddress<${typeParam}> | TransactionSigner<${typeParam}>`)
                .addImports('solanaAddresses', ['ProgramDerivedAddress'])
                .addImports('solanaSigners', ['TransactionSigner']);
        }
        if (account.isSigner === 'either') {
            return (0, common_1.fragment)(`Address<${typeParam}> | TransactionSigner<${typeParam}>`)
                .addImports('solanaAddresses', ['Address'])
                .addImports('solanaSigners', ['TransactionSigner']);
        }
        if (account.isSigner) {
            return (0, common_1.fragment)(`TransactionSigner<${typeParam}>`).addImports('solanaSigners', ['TransactionSigner']);
        }
    }
    if (!withSigners && account.isPda) {
        return (0, common_1.fragment)(`ProgramDerivedAddress<${typeParam}>`).addImports('solanaAddresses', ['ProgramDerivedAddress']);
    }
    return (0, common_1.fragment)(`Address<${typeParam}>`).addImports('solanaAddresses', [
        'Address',
    ]);
}
//# sourceMappingURL=instructionInputType.js.map